!
! Module: implot_mod
!
! Authors:  Clive Page - University of Leicester
!	    Extensively hacked by Ian Stewart from Clive's implot-1.6
!          
! Description: Plots EPIC/OM data using PGPLOT and WCS
! 
! Comments:
!	TODO:
!	- Implement \param{removeblack}.
!	- Re labels and radii: make the failure kinder if the requisite columns are not there. Ie check to see first, and eg if the radius column is not present, warn then make radii constant.
!	- Add 'constant' as a value for --radiusstyle.
!
! implot-1.6   Version with consistent internal/external numbering!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module implot_mod

  use dal
  use cal
  use param  
  use errorHandling

  use constants
  use wcs_aux
  use sscmisc
  use array_utils
  use fits_utils
  use mathutils
  use cal_aux
use dump

  implicit none 

  private 

  ! Command-line parameters:
  character(255) ::&
pgplotDevice,&
    imageSetName,&
    srcListSetName,&
    srcListSetTabName,&
    selectionExpression,&
    maxSourceNumSortExpression,&
    srcLabelExpression,&
    radiusExpression
  character(20) ::&
    srcListTabName,&
!    pgplotDevice,&
    radiusStyle,&
    maxSourceNumSortStyle,&
    srcLabelStyle,&
    gridStyle,&
    imageStyle,&
    imageTransferStyle,&
    contourStyle
  character(20), allocatable ::&
    userLabels(:)
  integer ::&
    colourMapId,&
    textClrIndex,&
    lineThicknessPixels,&
    numContours
  integer, allocatable ::&
    userSrcClrIndices(:)
  integer(int32) ::&
    userMaxSourceNum
  real ::&
    maxRadiusPixels,&
!userRadiusPixels,&
!*** now obsolete. Get rid of it as soon as you get rid of readParameters_old.
    textSize
  real(single) ::&
    zClipFractionLo,&
    zClipFractionHi,&
    fracBackground,&
    xClipFractionLo,&
    xClipFractionHi,&
    yClipFractionLo,&
    yClipFractionHi,&
    contourSeparation
  logical ::&
    withZclip,&
    withSrcList,&
    clipBlankBorder,&
    withEllipse,&
    withXYclip,&
    addUserSources,&
    withFrame
!  type(srcDataType), allocatable ::&
!    userSrcData(:)
  real(double), allocatable ::&
    userRaDeg(:), userDecDeg(:)

  ! Other globals:
  real ::&
    rhTextSize = 0.5,&
!    rhTextY = 1.05,&
    bottomTextSize = 0.7
  type(wcsDataType) ::&
    wcs
  character(len=8) ::&
    wcsType
  integer(int32) ::&
    instrumentId
  type(SpacecraftAttitudeType) ::&
    scAttitude, instrumAttitude
  real(double) ::&
    expStartTimeStamp,&
!    fovRadiusDeg,&
    pixelAreaArcsec
!,&
!    fovCtrXworld,&
!    fovCtrYworld
  integer ::&
    imageTransferId
  logical ::&
    deviceIsPrinter

  real, parameter ::&
    rhCharX = 1.0,&
    rhCharDy = 0.05,&
    rhCharXindent = 1.0,&
    rhCharYindent = -0.02

  public implot

contains 

  !---------------------------------------------------------------------
  subroutine implot
  !---------------------------------------------------------------------

    implicit none

    integer :: ierr, pgbeg
    logical :: imageIsInteger
    real(single), pointer :: image(:,:)
    real :: effMin, effMax
    integer :: minx, maxx, miny, maxy ! image pixel ranges
    logical, allocatable :: nonNullsArray(:,:)
!character(3) :: answerStr, pgDeviceRump
!integer :: answerLength, i, slashPosition
character(3) :: pgDeviceRump
integer :: i, slashPosition

    !-------------------------------------------------------------------
    ! Start

    call readParameters
    call loadInImage(image, imageIsInteger)
    allocate(nonNullsArray(size(image, 1), size(image, 2)))
    call fixUpNulls(image, nonNullsArray)

    if (withZclip) then
      effMin = valueAtBinFraction(real(image), zClipFractionLo, nonNullsArray)
      effMax = valueAtBinFraction(real(image), zClipFractionHi, nonNullsArray)
      ! function valueAtBinFraction() is in ssclib/array_utils.
    else
      effMin = minval(image)
      effMax = maxval(image)
      ! Don't need to mask with nonNullsArray - recall that fixUpNulls()
      ! replaces all nulls in the image by the minimum non-null value.
    end if

if (effMin .ge. effMax) call error('badZclipLevels','The minimum and &
&maximum image brightness truncation levels are the same.')

    ! set up PGPLOT.  Note use PGBEG not PGOPEN because latter
    ! does not get picked up from library on some installations
    ierr = pgbeg(0, trim(pgplotDevice), 1, 1)
    if(ierr /= 1)&
call error("badPGPLOTdevice", "Invalid PGPLOT device: "//trim(pgplotDevice))

    ! Avoid irritating phenomenon in which background and foreground
    ! colours swap for ps/cps output, except within the image:
    slashPosition = index(pgplotDevice, "/")
    pgDeviceRump = ucase(pgplotDevice(slashPosition+1:))
    if ((pgDeviceRump .eq. 'PS' .or. pgDeviceRump .eq. 'CPS')&
    .and. imageStyle .eq. 'image') then
      deviceIsPrinter = .true.

      if (textClrIndex .eq. 1) then
        textClrIndex = 0 
      else if (textClrIndex .eq. 0) then
        textClrIndex = 1
      end if

      do i = 1, parameterCount('userras')
        if(userSrcClrIndices(i) .eq. 1) then
          userSrcClrIndices(i) = 0
        else if(userSrcClrIndices(i) .eq. 0) then
          userSrcClrIndices(i) = 1
        end if
      end do
    end if ! device is postscript.

    call plotImage(image, effMin, effMax, minx, maxx, miny, maxy)
    pixelAreaArcsec = abs(wcs%X%pixelSizeWorld * wcs%Y%pixelSizeWorld&
    * 3600.0 * 3600.0)

    if(withSrcList) call addSources(minx, maxx, miny, maxy)
    call pgend
    deallocate(image, nonNullsArray)
    deallocate(userRaDeg, userDecDeg, userLabels)

end subroutine implot

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: numUserSrcs, i

    !-------------------------------------------------------------------
    ! Start

    imageSetName = stringParameter('set')
    pgplotDevice = stringParameter('device')
    gridStyle = stringParameter('gridstyle')
    imageStyle = stringParameter('imagestyle')
    imageTransferStyle = stringParameter('zscaletype')

    select case(imageTransferStyle)
    case('linear')
      imageTransferId = 0
    case('log')
      imageTransferId = 1
    case('sqrt')
      imageTransferId = 2
    case default
call error('badImageTransferId','The image transfer id (parameter &
&--zscaletype) is not recognized.')
    end select

    colourMapId = intParameter('colourmap')
    contourStyle = stringParameter('contourstyle')
    numContours = intParameter('ncontours')
    contourSeparation = realParameter('contourseparation')

    withZclip = booleanParameter('withzclip')
    zClipFractionLo = realParameter('zhistolo')
    zClipFractionHi = realParameter('zhistohi')

    if (withZclip .and. zClipFractionLo .ge. zClipFractionHi)&
call error('badZclipValues','The requested --zhistolo is greater than &
&or equal to --zhistohi.')

    clipBlankBorder = booleanParameter('trimborder')
    withXYclip = booleanParameter('withxyclip')
    xClipFractionLo = realParameter('xfraclo')
    xClipFractionHi = realParameter('xfrachi')
    yClipFractionLo = realParameter('yfraclo')
    yClipFractionHi = realParameter('yfrachi')

    withFrame = booleanParameter('withframe')

    withSrcList = booleanParameter('withsrclisttab')
    srcListSetTabName = stringParameter('srclisttab')
    selectionExpression = stringParameter('expression')

    maxSourceNumSortStyle = stringParameter('ncutsortstyle')
    maxSourceNumSortExpression = stringParameter('ncutsortexpression')
    userMaxSourceNum = intParameter('ncut')

    addUserSources = booleanParameter('addusersources')
    numUserSrcs = parameterCount('userras')
!print *, 'Number of user sources: ', numUserSrcs
!    allocate(userSrcData(numUserSrcs), userLabels(numUserSrcs)&
allocate(userRaDeg(numUserSrcs), userDecDeg(numUserSrcs), userLabels(numUserSrcs)&
    , userSrcClrIndices(numUserSrcs))
    do i = 0, numUserSrcs - 1
!      userSrcData(i + 1)%raDeg   = realParameter('userras',      i)
!      userSrcData(i + 1)%decDeg  = realParameter('userdecs',     i)
userRaDeg(i + 1)  = realParameter('userras',      i)
userDecDeg(i + 1) = realParameter('userdecs',     i)
      userLabels(i + 1)          = stringParameter('userlabels', i)
      userSrcClrIndices(i + 1)   = intParameter('usercolours',   i)
    end do

    radiusStyle = stringParameter('radiusstyle')
    radiusExpression = stringParameter('radiusexpression')
!    userRadiusPixels = realParameter('radius')
maxRadiusPixels = realParameter('maxradius')
    withEllipse = booleanParameter('withellipse')
    fracBackground = realParameter('bkgfraction')
!withEllipse = .false.
!fracBackground = 1.0

    srcLabelStyle = stringParameter('labelstyle')
    srcLabelExpression = stringParameter('labelexpression')
    textClrIndex = intParameter('labelcolour')
    textSize = realParameter('labelsize')
    lineThicknessPixels = intParameter('thickness')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine loadInImage(image, imageIsInteger)
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: image(:,:)
    logical, intent(out) :: imageIsInteger

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Reading image.')

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)

    call loadImage(inArray, image, imageIsInteger)
!    call getImageWcsInfo(set, inArray, wcs)
call getWcsInfo(set, inArray, wcs)
    call easifyWcsType(wcs%X%ctype, wcs%Y%ctype, wcsType)

    if (abs(wcs%X%pixelSizeWorld) .ne. abs(wcs%Y%pixelSizeWorld)) then
!print *, wcs%X%pixelSizeWorld, wcs%Y%pixelSizeWorld
!call error('nonSquarePixels', 'Sorry, only square pixels are at present supported.')
!**** Have to insist upon square pixels until I get this ellipse projection sorted out.
    end if

    call release(set)

  end subroutine loadInImage

  !---------------------------------------------------------------------
  subroutine loadImage(inArray, image, imageIsInteger)
  !---------------------------------------------------------------------
  ! Adapted from asmooth-2.5

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), dimension(:,:), pointer :: image
    logical, intent(out) :: imageIsInteger

    integer(int32), dimension(:), pointer ::&
      dimVector
    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64
    integer(int16) :: xsize, ysize

    !-------------------------------------------------------------------
    ! Start:

    imageIsInteger = .false. ! default

    dimVector => dimensions(inArray)
    xsize = dimVector(0)
    ysize = dimVector(1)

    allocate(image(xsize, ysize))

    if (dataType(inArray) .eq. INTEGER8) then
      arrayDataInt8 => int8Array2Data(inArray)
      image = real(arrayDataInt8, single)
      imageIsInteger = .true.
    else if (dataType(inArray) .eq. INTEGER16) then
      arrayDataInt16 => int16Array2Data(inArray)
      image = real(arrayDataInt16, single)
      imageIsInteger = .true.
    else if (dataType(inArray) .eq. INTEGER32) then
      arrayDataInt32 => int32Array2Data(inArray)
      image = real(arrayDataInt32, single)
      imageIsInteger = .true.
    else if (dataType(inArray) .eq. REAL32) then
      arrayDataReal32 => real32Array2Data(inArray)
      image = arrayDataReal32
    else ! this gives a DAL error if data type is none of the above.
      arrayDataReal64 => real64Array2Data(inArray)
      image = real(arrayDataReal64, single)
    end if

  end subroutine loadImage

  !---------------------------------------------------------------------
  subroutine plotImage(image, effMin, effMax, minx, maxx, miny, maxy)
  !---------------------------------------------------------------------

    use metatask_mod ! to get task name and version number.

    implicit none

    real(single), intent(in) :: image(:,:)
    real, intent(in) :: effMin, effMax   ! effective range of image
    integer, intent(out) :: minx, maxx, miny, maxy

    integer :: status, xsize, ysize
    real :: blc(2)             ! bottom left corner
    real(double) :: cache(4,0:256)
    integer                      :: ci(7) = -1         ! colours for grid
    double precision             :: dparam(50)=0.0d0 ! parameters for PGWCSL
    integer                      :: gcode(2) = 1     ! GCODE arg for PGCRVL
    integer                      :: ic
    character(len=80)            :: idents(3)     ! annotation for axes, title
    integer                      :: ierr          
    type(ArrayT)                 :: imageArray
    type(dataSetT)               :: imageSet
    integer                      :: ksl
    integer                      :: labctl       ! label control on pgsbox
    character(len=80)            :: legend
    real, parameter              :: tr(6) = (/ 0.0, 1.0, 0.0, 0.0, 0.0, 1.0 /)
    real                         :: trc(2)       ! top right corner
    integer                      :: wcsThingie(4)
    type(TaskInfo) :: info ! type from metatask_mod
    external pgwcsl
    character(len=8) :: ctype(2)
    integer :: minx_temp, miny_temp, i
    real, allocatable :: contours(:)

    !-------------------------------------------------------------------
    ! Start

    info = getTaskInfo() ! gets task name and version number. Routine is in
    ! metatask_mod.

    xsize = size(image, 1)
    ysize = size(image, 2)

    if (clipBlankBorder) then
      call maskArrayLimits((image .ne. 0.0), minx, maxx, miny, maxy, status)
if (status .eq. -1) call error('imageAllZero','There are no non-zero &
&pixels in the image.')
if (status .eq. -2) call error('sizelessImageArray','The image array &
&has zero size.')

    else
      minx = 1
      maxx = xsize
      miny = 1
      maxy = ysize
    end if

    if (withXYclip) then
      minx_temp = nint(minx * (1.0 - xClipFractionLo) + maxx * xClipFractionLo)
      maxx = nint(minx * (1.0 - xClipFractionHi) + maxx * xClipFractionHi)
      minx = minx_temp
      miny_temp = nint(miny * (1.0 - yClipFractionLo) + maxy * yClipFractionLo)
      maxy = nint(miny * (1.0 - yClipFractionHi) + maxy * yClipFractionHi)
      miny = miny_temp
if (minx .ge. maxx) call error('badXclipValues','The requested --xfraclo &
&is greater than or equal to --xfrachi.')
if (miny .ge. maxy) call error('badYclipValues','The requested --yfraclo &
&is greater than or equal to --yfrachi.')
    end if

    if (withFrame) then
      call pgsvp(0.1, 0.9, 0.13, 0.93) 
    else
      call pgsvp(0.0, 1.0, 0.0, 1.0) 
    end if
    call pgwnad(minx-0.5, maxx+0.5, miny-0.5, maxy+0.5)
    call pgsch(bottomTextSize)

!y = fracVal(real(miny), real(maxy), 0.5)
!x = fracVal(real(minx), real(maxx), 0.1)
!call pgsci(4)
!call pgtext(x, y, '4')
!x = fracVal(real(minx), real(maxx), 0.2)
!call pgsci(5)
!call pgtext(x, y, '5')
!x = fracVal(real(minx), real(maxx), 0.3)
!call pgsci(6)
!call pgtext(x, y, '6')

!return

    select case(imageStyle)
    case('contour')
      if (contourStyle .eq. 'separation') then
        numContours = 1 + max(0, floor((effMax - effMin) / contourSeparation))
        if (numContours <= 1) call warning('contourSeparationTooLarge'&
          ,'The specified contour separation results in <= 1 contour.')

      else if (contourStyle .ne. 'number') then
call error('badContourStyle','The contour style is not recognized.')
      end if

      allocate(contours(numContours))

      if (numContours == 1) then
        contours(1) = effMin

      else if (contourStyle .eq. 'separation') then
        do i = 1, numContours
          contours(i) = effMin + (i -1) * contourSeparation
        end do

      else ! contourStyle must .eq. 'number':
        do i = 1, numContours
          contours(i) = fracVal(effMin, effMax, (i - 1) / real(numContours))
        end do
      end if

do i = 1, numContours
  print *, 'i, contours(i):', i, contours(i)
end do

      call pgcons(real(image), xsize, ysize, minx, maxx, miny, maxy&
        , contours, numContours, tr)

      deallocate(contours)

    case('image')
      call set_palette(colourMapId)
      call pgsitf(imageTransferId)
      call pgimag(real(image), xsize, ysize, minx, maxx, miny, maxy&
        , effMin, effMax, tr)
    case default
call error('badImageType', 'The output image type (ie, colour-coded &
&or contour) is not recognized.')
    end select

    write(errStr, '(a, es10.3)') 'Highest image value shown: ', effMax
    call message(AppMsg, Verbose)
    write(errStr, '(a, es10.3)') 'Lowest image value shown: ', effMin
    call message(AppMsg, Verbose)

    if (withFrame) then
      ! plot title consisting of input file-name at centre of top
      ksl = index(imageSetName, "/", .true.)  ! points to last slash
      call pgmtxt("T", 1.4, 0.5, 0.5, trim(imageSetName(ksl+1:)))
      write(legend, "(a)") trim(nameOf(info))//' v'//trim(versionOf(info))
      call pgmtxt("T", 1.2, 1.0, 0.0, legend)
      call pgscf(2)                 ! better quality Roman font

      if(gridStyle .ne. 'none') then
        call pgbox('BC', 0.0, 0, 'BC', 0.0, 0)
        call pglab(' ', wcs%Y%ctype, ' ')
      else
        call pgbox('BCINT', 0.0, 0, 'BCINT', 0.0, 0)
        call pglab('X', 'Y', " ")
      end if
    end if

    imageSet = dataSet(imageSetName, READ)
    imageArray = array(imageSet, 0, READ)
    if (withFrame) call plotRHinfo(minx, maxx, miny, maxy, imageSet, imageArray)
    call release(imageSet)

    ! plot coordinate grid over the top of the image if requested
    if(gridStyle .ne. 'none') then
      idents(1) = wcs%x%ctype
      idents(2) = " "
      idents(3) = " "
      ic        = -1
      dparam(4)     = wcs%x%refPixels 
      dparam(5)     = wcs%y%refPixels 
      dparam(6:9)   = (/1.0d0, 0.0d0, 0.0d0, 1.0d0/)
      dparam(10) = wcs%X%pixelSizeWorld
      dparam(11) = wcs%Y%pixelSizeWorld
      dparam(20) = wcs%X%refWorld
      dparam(21) = wcs%Y%refWorld
      dparam(22)    = 999d0
      dparam(23)    = 999d0
      dparam(31)    = 0.0d0
      dparam(32) = cos(PiOn180 * wcs%Y%refWorld)&
                 / sin(PiOn180 * wcs%Y%refWorld)
      select case(gridStyle)
      case('ticks')
        gcode = 1
      case('grid')
        gcode = 2
      case default
call error('badGridStyle', 'The grid style is not recognized.')
      end select
      blc(1) = minx - 0.5
      blc(2) = miny - 0.5
      trc(1) = maxx + 0.5
      trc(2) = maxy + 0.5
!print *, 'minx, maxx, miny, maxy: ', minx, maxx, miny, maxy
!print *, 'wcs%x%refPixels, wcs%y%refPixels: ', wcs%x%refPixels, wcs%y%refPixels
!print *, 'wcs%X%refWorld, wcs%Y%refWorld: ', wcs%X%refWorld, wcs%Y%refWorld
!print *, 'wcs%X%pixelSizeWorld, wcs%Y%pixelSizeWorld: ', wcs%X%pixelSizeWorld, wcs%Y%pixelSizeWorld
      if(gridStyle .eq. 'grid') ci(1:2) = 4       ! overlay with blue lines
      labctl = 21

      ctype(1) = wcs%X%ctype
      ctype(2) = wcs%Y%ctype
      call pgsbox(blc, trc, idents, (/"G","E"/), labctl, 0, ci, gcode&
      , -2.0d0, 0, 0.0d0, 0, 0.0d0, .false., pgwcsl, 16, 4, 50, ctype&
      , wcsThingie, dparam, 256, ic, cache, ierr)

      if(ierr /= 0) then
call error("pgcrvlErr", "error in pgsbox")
      end if
    end if

    if (withFrame .and. imageStyle .eq. 'image') then
      call pgscf(2)
      call pgwedg("LI", 4.9, 2.3, effMin, effMax, " ")
      call pgsch(0.5)
      call pgmtxt("T", 1.2, -0.07, 1.0, 'cts/pixel')
      call pgsch(1.0)
      call pgiden
    end if

  end subroutine plotImage

  !-----------------------------------------------------------------------
  subroutine plotRHinfo(minx, maxx, miny, maxy, imageSet, imageArray)
  !-----------------------------------------------------------------------
  ! shows FITS headers at top of plot

!    use fitsutils_mod ! contains getStringKwd() etc and intToStr()

    implicit none 

    integer,        intent(in) :: minx, maxx, miny, maxy
    ! world coordinate limits
    type(DataSetT), intent(in) :: imageSet
    type(ArrayT),   intent(in) :: imageArray

    integer :: font, status
    character(len=128) :: string
    real :: x, y, dy, csize, xoffset, yoffset

    !-------------------------------------------------------------------
    ! Start:

    x = rhCharX
    y = 0.95
    dy = rhCharDy
xoffset = rhCharXindent
yoffset = rhCharYindent

    call pgqcf(font)
    call pgscf(1)
    call pgqch(csize)
    call pgsch(rhTextSize)

    call pgmtxt("RV", x, y, 0.0, 'Instrument:')
    call getStringKwd('INSTRUME', imageSet, imageArray, string, status)
    if(status .eq. 0) call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(string))

    call getStringKwd('FILTER', imageSet, imageArray, string, status)
    y = y - dy
    call pgmtxt("RV", x, y, 0.0, 'Filter:')
    if(status .eq. 0) call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(string))

    call getStringKwd('OBJECT', imageSet, imageArray, string, status)
    y = y - dy
    call pgmtxt("RV", x, y, 0.0, 'Object:')
    if(status .eq. 0) call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(string))

    call getStringKwd('OBSERVER', imageSet, imageArray, string, status)
    y = y - dy
    call pgmtxt("RV", x, y, 0.0, 'Observer:')
    if(status .eq. 0) call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(string))

    call getStringKwd('DATE-OBS', imageSet, imageArray, string, status)
    y = y - dy
    call pgmtxt("RV", x, y, 0.0, 'DATE-OBS:')
    if(status .eq. 0) call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, string(1:16))

    call getStringKwd('EXPOSURE', imageSet, imageArray, string, status)
    y = y - dy
    call pgmtxt("RV", x, y, 0.0, 'Exposure:')
    if(len_trim(string) > 0 .and. status .eq. 0) then
      string = trim(string) // ' sec'
      call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(string)) 
    end if


    write(string, "('pixels (', i0, ':', i0, ',', i0, ':', i0, ')')") &
      minx, maxx, miny, maxy
    y = y - 1.5 * dy
    call pgmtxt("RV", x, y, 0.0, 'Image size:')
    call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(string))


    y = y - 1.5 * dy
    call pgmtxt("RV", x, y, 0.0, 'Parameters:')
    y = y - 0.5 * dy
    call pgmtxt("RV", x+xoffset, y, 0.0, 'scaletype = '&
    //trim(imageTransferStyle))
    y = y - 0.5 * dy
    write(string, "('zlo = ', f8.6)") zClipFractionLo
    call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
    y = y - 0.5 * dy
    write(string, "('zhi = ', f8.6)") zClipFractionHi
    call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
    if (withSrcList) then
      y = y - 0.5 * dy
!      write(string, "('radius = ', f6.2)") userRadiusPixels
      write(string, "('radius = ', f6.2)") maxRadiusPixels
      call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
    end if
    if (withXYclip) then
      y = y - 0.5 * dy
      write(string, "('xlo = ', f6.4)") xClipFractionLo
      call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
      y = y - 0.5 * dy
      write(string, "('xhi = ', f6.4)") xClipFractionHi
      call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
      y = y - 0.5 * dy
      write(string, "('ylo = ', f6.4)") yClipFractionLo
      call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
      y = y - 0.5 * dy
      write(string, "('yhi = ', f6.4)") yClipFractionHi
      call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
    end if
    if (addUserSources) then
!      if (size(userSrcData) .le. 1) then
      if (size(userRaDeg) .le. 1) then
        y = y - 0.5 * dy
!        write(string, "('user RA (deg) = ', f10.5)") userSrcData(1)%raDeg
        write(string, "('user RA (deg) = ', f10.5)") userRaDeg(1)
        call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
        y = y - 0.5 * dy
!        write(string, "('user dec (deg) = ', f10.5)") userSrcData(1)%decDeg
        write(string, "('user dec (deg) = ', f10.5)") userDecDeg(1)
        call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))
      else
        y = y - dy
!        string = intToStr(size(userSrcData))
        string = intToStr(size(userRaDeg))
        string = trim(string)//' user sources added.'
        call pgmtxt("RV", x, y, 0.0, trim(string))
      end if
    end if
    y = y - 0.5 * dy
    write(string, "('colourmap = ', i3)") colourMapId
    call pgmtxt("RV", x+xoffset, y, 0.0, trim(string))

    call pgscf(font)
    call pgsch(csize)

  end subroutine plotRHinfo

  !----------------------------------------------------------------------
  subroutine addSources(minx, maxx, miny, maxy)
  !----------------------------------------------------------------------

    use datafilter
!    use fitsutils_mod
    use sort_mod
use source_cutouts

    implicit none
!sortExpr
    integer, intent(in) :: minx, maxx, miny, maxy ! image pixel ranges

    character(255) ::&
      text
    character(20) ::&
      radiusColName,&
      sortColName, labelsColName
    type(dataSetT) ::&
      sourceSet, imageSet
    type(TableT) ::&
      sourceTab
    integer(int32) ::&
      numFilteredRows,&
      numLastSource,&
      src,&
      numOutside,&
      numShown,&
      ksl!!!, int32_dummy
    real, pointer ::&
!      shortAxes(:), longAxes(:), anglesDeg(:)
xAmp(:), yAmp(:), phase(:), srcXpixels(:), srcYpixels(:)
logical, pointer :: srcIsValid(:)
    real, allocatable ::&
      tempRadii(:)
!    real(single), pointer ::&
!      counts(:), bkgFlux(:)
    integer(int32), pointer ::&
      indices(:),&
      tempIndices(:),&
!      indicesFirstIndexZero(:),&
      numbers(:),&
!      tempNumbers(:),&
      labelsFirstIndexZero(:)
    type(ColumnT) ::&
      radiusCol
!,&
!      sortCol
    real(single) ::&
      maxRadiusDataValue
    real(double), allocatable ::&
      sortColData(:),&
      radiusColData(:),&
      raDeg(:),&
      decDeg(:)
    real(double) ::&
      cel(10),&
      prj(0:20),&
      lin(20),&
      world(2),&
      imgcrd(2),&
      pixcrd(2),&
      phi,&
      theta
    real ::&
      x,&
      xOffset, yOffset,&
      y,&
      dy,&
      csize,&
      minRadiusFactor = 0.20,&
      maxRadiusFactor = 1.0
    integer ::&
      wcsThingie(0:3),&
      colsave,&
      fontsave,&
      thicksave,&
      j,&
      ierr,&
      posOfColon
    logical ::&
      allRowsNull
    real(double) :: crval(2)
    character(len=8) :: ctype(2)
integer :: i
  type(wcsDataType) :: srcWcs
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Open source list.')

!........................................................................
! Load and filter source list:

    posOfColon = index(srcListSetTabName, ':')
    if (posOfColon .le. 0) then ! no colon in the set-tab name. User
    ! has forgotten to include the table name.
      call warning('missingTabName', "No colon+table name found after &
        &the set name. I'm going to use the first table in the set.")
      srcListSetName = srcListSetTabName
      srcListTabName = name(table(dataSet(srcListSetName, READ), 0))

    else
      srcListSetName = srcListSetTabName(:posOfColon - 1)
      srcListTabName = srcListSetTabName(posOfColon + 1:)
    end if
!*** use splitSetTabName in fits_utils

    sourceSet = clone(srcListSetName, 'tempset.ds', TEMP)
    sourceTab = table(sourceSet, srcListTabName, MODIFY)

    call message(AppMsg, Verbose, 'Selection of sources.')

    call initFilter(selectionExpression)
    numFilteredRows = doFilter(sourceTab)

    if (numFilteredRows .le. 0)&
call error('noSourcesSelected'&
, 'No sources passed the specified filtering expression.')

!........................................................................
! Calculate radii of source circles:

    call message(AppMsg, Verbose, 'Calculating radii.')

!    allocate(shortAxes(numFilteredRows), longAxes(numFilteredRows)&
!    , anglesDeg(numFilteredRows))

    allocate(raDeg(numFilteredRows), decDeg(numFilteredRows))
    call readColData(sourceTab, 'RA',  raDeg)
    call readColData(sourceTab, 'DEC', decDeg)

!    anglesDeg = 0.0 ! default

    if (radiusStyle .eq. 'expr') then
allocate(xAmp(numFilteredRows), yAmp(numFilteredRows)&
, phase(numFilteredRows), srcIsValid(numFilteredRows), srcXpixels(numFilteredRows), srcYpixels(numFilteredRows))

phase = PiOverTwo ! default

radiusColName = radiusExpression
radiusCol = column(sourceTab, radiusColName)
!*** eventually hope to have a routine in selectlib which will do arithmetic on columns. For now, column() will obviously fail with an error if radiusExpression is more than a simple column name.

      call maxNonNullValue(sourceTab, radiusColName, maxRadiusDataValue&
      , allRowsNull)

      if (allRowsNull)&
call error('allRadiusRowsNull', 'All the rows specified by &
&--radiusexpression are null-valued.')

allocate(radiusColData(numFilteredRows))
      call readColData(radiusCol, radiusColData)
!      longAxes = 1.0 + 0.5 * log10(radiusColData/maxRadiusDataValue)
yAmp = 1.0 + 0.5 * log10(radiusColData/maxRadiusDataValue)
      deallocate(radiusColData)
!      where(longAxes .gt. maxRadiusFactor) longAxes = maxRadiusFactor
where(yAmp .gt. maxRadiusFactor) yAmp = maxRadiusFactor
!      where(longAxes .lt. minRadiusFactor) longAxes = minRadiusFactor
where(yAmp .lt. minRadiusFactor) yAmp = minRadiusFactor
!      longAxes = userRadiusPixels * longAxes
!longAxes = maxRadiusPixels * longAxes
yAmp = maxRadiusPixels * yAmp

!      shortAxes = longAxes
xAmp = yAmp
srcIsValid = .true.

    else if (radiusStyle .eq. 'psf') then
!      allocate(counts(numFilteredRows), bkgFlux(numFilteredRows))
!      call readColData(column(sourceTab, 'SCTS'),  counts)
!      call readColData(column(sourceTab, 'BG_MAP'), bkgFlux)
!      bkgFlux = bkgFlux / pixelAreaArcsec
!      call getPsfRadii(raDeg, decDeg, counts, bkgFlux, shortAxes, longAxes&
!      , anglesDeg)
!      deallocate(counts, bkgFlux)

    imageSet = dataSet(imageSetName, READ)
    call getCalInfo(imageSet, instrumentId, expStartTimeStamp, scAttitude&
      , instrumAttitude) ! sets CAL instrument state.

    if (instrumentId .ne. EMOS1&
    .and. instrumentId .ne. EMOS2&
    .and. instrumentId .ne. EPN) then
call error('badInstrum', 'Instrument '//trim(instId(instrumentId))//&
&' not supported.')
    end if

    call release(imageSet)

srcWcs%X%refPixels       = int32Attribute(sourceSet, 'TCRPIX1')
srcWcs%Y%refPixels       = int32Attribute(sourceSet, 'TCRPIX2')
srcWcs%X%pixelSizeWorld = real32Attribute(sourceSet, 'TCDELT1')
srcWcs%Y%pixelSizeWorld = real32Attribute(sourceSet, 'TCDELT2')
srcWcs%X%refWorld       = real32Attribute(sourceSet, 'TCRVAL1')
srcWcs%Y%refWorld       = real32Attribute(sourceSet, 'TCRVAL2')
srcWcs%X%ctype          = stringAttribute(sourceSet, 'TCTYPE1')
srcWcs%Y%ctype          = stringAttribute(sourceSet, 'TCTYPE2')

call getSourceCutouts(sourceTab, wcs, srcWcs, scAttitude&
, expStartTimeStamp, fracBackground, 'RA', 'DEC'&
, 'SCTS', 'BG_MAP', srcXpixels, srcYpixels&
, xAmp, yAmp, phase, srcIsValid)

if (testFlag) then
print *, '** now back out in implot.'
do src = 1, size(raDeg)
  print *, 'src, xAmp(src), yAmp(src), phase(src): '&
  , src, xAmp(src), yAmp(src), phase(src)
end do
print *, ''
end if

!*** srcXpixels, srcYpixels not used at present.

!    else if (radiusStyle .eq. 'user') then
!      longAxes = userRadiusPixels
!      shortAxes = longAxes
    end if

!........................................................................
! If a limit on the number of sources to be plotted is desired, the sources
! must first be sorted. This is accomplished next:

!print *, 'maxSourceNumSortStyle=', maxSourceNumSortStyle

    if (maxSourceNumSortStyle .ne. 'none')&
      call message(AppMsg, Verbose, 'Sorting to prepare for max source &
      &number cut.')

    allocate(indices(numFilteredRows)&
    , tempRadii(numFilteredRows))

    select case(maxSourceNumSortStyle)
    case('radiusup')
!      tempRadii = longAxes
tempRadii = yAmp
      tempIndices => indicize(numFilteredRows)
      call quick_sort(tempRadii, tempIndices)
      do i = 1, size(tempIndices)
        indices(i) = tempIndices(i)
      end do
      deallocate(tempIndices)
!print *, 'deallocated tempIndices'

    case('radiusdown')
!      tempRadii = -longAxes
tempRadii = -yAmp
      tempIndices => indicize(numFilteredRows)
      call quick_sort(tempRadii, tempIndices)
      do i = 1, size(tempIndices)
        indices(i) = tempIndices(i)
      end do
      deallocate(tempIndices)
!print *, 'deallocated tempIndices'

    case('expr')
sortColName = maxSourceNumSortExpression
!*** eventually hope to have a routine in selectlib which will do arithmetic on columns. For now, column() will obviously fail with an error if maxSourceNumSortExpression is more than a simple column name.

!      sortCol = column(sourceTab, sortColName)
allocate(sortColData(numFilteredRows))
      call readColData(sourceTab, sortColName, sortColData)
sortColData = -sortColData
      tempIndices => indicize(numFilteredRows)
      call quick_sort(sortColData, tempIndices)
      do i = 1, size(tempIndices)
!        indices(tempIndices(i)) = i
        indices(i) = tempIndices(i)
      end do
      deallocate(sortColData)

!call readColData(sortCol, sortColData)
!do i = 1, size(tempIndices)
!print *, 'i, tempIndices(i), indices(i): ', i, tempIndices(i), indices(i)
!print *, '    sortColData(i), sortColData(tempIndices(i)), sortColData(indices(i)): '&
!, sortColData(i-1), sortColData(tempIndices(i)-1), sortColData(indices(i)-1)
!end do
!deallocate(sortColData)
!print *, 'deallocated sortColData'

      deallocate(tempIndices)
!print *, 'deallocated tempIndices'

    case('rownumber')
      tempIndices => indicize(numFilteredRows)
      indices = tempIndices
      deallocate(tempIndices)
!print *, 'deallocated tempIndices'

    case('none') ! no action
      tempIndices => indicize(numFilteredRows)
      indices = tempIndices
      deallocate(tempIndices)
!print *, 'deallocated tempIndices'
    case default
call error('badNcutSortStyle', 'The value of --ncutsortstyle is &
&not recognized.')
    end select

!........................................................................
! Allocating source labels:

    if (srcLabelStyle .ne. 'none')&
      call message(AppMsg, Verbose, 'Allocating source labels.')

    allocate(numbers(numFilteredRows))

    select case(srcLabelStyle)
    case('radiusup')
!      tempRadii = longAxes
tempRadii = yAmp
      tempIndices => indicize(numFilteredRows)
      call quick_sort(tempRadii, tempIndices)
      do i = 1, size(tempIndices)
        numbers(tempIndices(i)) = i
      end do
      deallocate(tempIndices)
!print *, 'deallocated tempIndices'

    case('radiusdown')
!      tempRadii = -longAxes
tempRadii = -yAmp
      tempIndices => indicize(numFilteredRows)
      call quick_sort(tempRadii, tempIndices)
      do i = 1, size(tempIndices)
        numbers(tempIndices(i)) = i
      end do
      deallocate(tempIndices)
!print *, 'deallocated tempIndices'

    case('expr')
labelsColName = srcLabelExpression
!*** eventually hope to have a routine in selectlib which will do arithmetic on columns. For now, column() will obviously fail with an error if srcLabelExpression is more than a simple column name.

allocate(labelsFirstIndexZero(numFilteredRows))
      call readColData(sourceTab, labelsColName, labelsFirstIndexZero)
!*** have to change this to allow string-valued columns (and labels).
      numbers = labelsFirstIndexZero
      deallocate(labelsFirstIndexZero)
!print *, 'deallocated labelsFirstIndexZero'

    case('rownumber')
      tempIndices => indicize(numFilteredRows)
      numbers = tempIndices
      deallocate(tempIndices)
!print *, 'deallocated tempIndices'

    case('none') ! no action
    case default
call error('badLabelsStyle', 'The label style is not recognized.')
    end select
    deallocate(tempRadii)
!print *, 'deallocated tempRadii'

!........................................................................
! Preparing for plot:

    wcsThingie = (/ 0, 1, 2, 0/)
    cel(1:4)  = (/ wcs%X%refWorld, wcs%Y%refWorld, 999.0d0, 999.0d0 /)
    cel(5:10) = 0.0d0
    prj  =  0.0d0
    lin(1:11) = (/ 0.0d0, 2.0d0, 2.0d0, dble(wcs%X%refPixels)&
    , dble(wcs%Y%refPixels)&
    , 1.0d0, 0.0d0, 0.0d0, 1.0d0, wcs%X%pixelSizeWorld, wcs%Y%pixelSizeWorld /)

    crval(1) = wcs%X%refWorld
    crval(2) = wcs%Y%refWorld
    ctype(1) = wcs%X%ctype
    ctype(2) = wcs%Y%ctype

    call pgqci(colsave)
    call pgsci(textClrIndex)
    call pgqcf(fontsave)
    call pgscf(1)
    call pgsfs(2)    ! for hollow circles
    call pgqlw(thicksave)
    call pgslw(lineThicknessPixels)
    call pgqch(csize)
    call pgsch(textSize)

!........................................................................
! Plot the source circles + labels:

    call message(AppMsg, Verbose, 'Plotting sources.')

    if (maxSourceNumSortStyle .eq. 'none') then
      numLastSource = numFilteredRows
    else
      numLastSource = min(userMaxSourceNum, numFilteredRows)
    end if

!print *, 'numFilteredRows', numFilteredRows
!print *, 'numLastSource', numLastSource
    numOutside = 0
    numShown = 0
    do src = 1, numLastSource
      j = indices(src)

      ! transfer ra/dec to wcs
!      world(1) = raDeg(j-1)
!      world(2) = decDeg(j-1)
world(1) = raDeg(j)
world(2) = decDeg(j)
      call wcsfwd(ctype, wcsThingie, world, crval, cel, phi, theta, prj, imgcrd&
      , lin, pixcrd, ierr)

      x = real(pixcrd(1)) 
      y = real(pixcrd(2)) 

      if(x < minx .or. x > maxx .or. y < miny .or. y > maxy) then
        numOutside = numOutside + 1

      else
        numShown = numShown + 1

        if (radiusStyle .eq. 'psf' .and. withEllipse) then
!          call drawSrcEllipse_20020827(x, y, shortAxes(j), longAxes(j), anglesDeg(j))
call drawSrcEllipse(x, y, xAmp(j), yAmp(j), phase(j))
if (testFlag) then
  print *, 'j, xAmp(j), yAmp(j), phase(j): '&
  , j, xAmp(j), yAmp(j), phase(j)
end if

        else ! draw hollow circle:
!          call pgcirc(x, y, longAxes(j))
call pgcirc(x, y, yAmp(j))
        end if

        if(srcLabelStyle .ne. 'none') then
!          call pgptxt(x + longAxes(j) + 1.0, y, 0.0, 0.0&
call pgptxt(x + yAmp(j) + 1.0, y, 0.0, 0.0&
          , trim(intToStr(numbers(j))))
!**** not quite able to print alphanumeric labels yet.
        end if
      end if
    end do

!if (testFlag) call pgpt1(346.1, 313.2, 2)

    if(numOutside > 0) then
      write(errStr, '(i6,1x,a)') numOutside, 'source positions outside image.'
      call warning('sourcesOutside')
    end if

    if (addUserSources) then
!      do src = 1, size(userSrcData)
      do src = 1, size(userRaDeg)
!        world(1) = userSrcData(src)%raDeg
!        world(2) = userSrcData(src)%decDeg
world(1) = userRaDeg(src)
world(2) = userDecDeg(src)
        call wcsfwd(ctype, wcsThingie, world, crval, cel, phi, theta&
        , prj, imgcrd, lin, pixcrd, ierr)

        x = real(pixcrd(1)) 
        y = real(pixcrd(2)) 
        call pgsci(userSrcClrIndices(src))
        call pgpt1(x, y, 2)
        call pgptxt(x + 1.0, y + 1.0, 0.0, 0.0, trim(userLabels(src)))
      end do
    end if

!........................................................................
! Add notations to base and side of plot:

    call message(AppMsg, Verbose, 'Remaining inscriptions.')

    if (withFrame) then
      ! Annotate plot at bottom with filename:

      call pgslw(thicksave)
      call pgsci(1)
      call pgsch(rhTextSize)

      x = rhCharX
      y = 0.25
      dy = rhCharDy
      xoffset = rhCharXindent
      yoffset = rhCharYindent

      call pgmtxt("RV", x, y, 0.0, "Sources from:")
      ksl = index(srcListSetTabName, "/", .true.)
      call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(srcListSetTabName(ksl+1:)))

      if (srcLabelStyle .ne. 'none') then
        y = y - dy
        call pgmtxt("RV", x, y, 0.0, "Numerals from:")
        select case(srcLabelStyle)
        case('radiusup')
          text = 'increasing radius.'
        case('radiusdown')
          text = 'decreasing radius.'
        case('expr')
          text = 'column '//trim(labelsColName)//'.'
        case('rownumber')
          text = 'row numbers.'
        end select

        call pgmtxt("RV", x+xoffset, y+yoffset, 0.0, trim(text))
      end if

      call pgsch(bottomTextSize)

      call pgmtxt("B", 4.4, -0.1, 0.0&
     , "Selection expr: '"//trim(selectionExpression)//"'")

      select case(maxSourceNumSortStyle)
      case('none')
        write(text, "(a, i0, a)") "Number of sources: all "&
        , numLastSource, " selected."

      case('radiusup')
        write(text, "(a, i0, a)") "Number of sources: first ", numLastSource&
        , " in increasing order of "//trim(radiusExpression)//"."

      case('radiusdown')
        write(text, "(a, i0, a)") "Number of sources: first ", numLastSource&
        , " in decreasing order of "//trim(radiusExpression)//"."

      case('expr')
        write(text, "(a, i0, a)") "Number of sources: first ", numLastSource&
        , " in decreasing order of "//trim(maxSourceNumSortExpression)//"."

      case('rownumber')
        write(text, "(a, i0, a)") "Number of sources: first ", numLastSource&
        , " in increasing order of row number."
      end select
!      if (maxSourceNumSortStyle .ne. 'none') then
!        write(text, "(a, i0, a)") "Number of sources: first ", numLastSource&
!        , " in order of "//trim(maxSourceNumSortExpression)//"."
!      else
!
!        write(text, "(a, i0, a)") "Number of sources: all ", numLastSource, " selected."
!      end if
      call pgmtxt("B", 5.4, -0.1, 0.0, text)

      if (radiusStyle .eq. 'expr') then
        text = "Radii: scaled from expression "//trim(radiusExpression)//"."
      else if (radiusStyle .eq. 'column') then
        text = "Radii: scaled from column "//trim(radiusColName)//"."
      else if (radiusStyle .eq. 'psf') then
        text = "Radii: scaled from PSF spread."
      else
        text = "Radii: fixed."
      end if
      call pgmtxt("B", 6.4, -0.1, 0.0, text)
    end if

    call message(AppMsg, Verbose, 'Clean up.')

    call pgsci(colsave)
    call pgscf(fontsave)
    call pgsch(csize)

    deallocate(indices, numbers)
!    deallocate(shortAxes, longAxes, anglesDeg)
deallocate(xAmp, yAmp, phase, srcIsValid, srcXpixels, srcYpixels)
    deallocate(raDeg)
    deallocate(decDeg)

    call message(AppMsg, Verbose, 'Release source set.')

    call release(sourceSet)

  end subroutine addSources

  !---------------------------------------------------------------------
  subroutine drawSrcEllipse(x, y, xAmp, yAmp, phase)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: x, y, xAmp, yAmp, phase

    real :: xVals(20), yVals(20)

    !-------------------------------------------------------------------
    ! Start

    if (xAmp .gt. 0.0 .and. yAmp .gt. 0.0) then
      call calcEllipseFromPhase(xAmp, yAmp, phase, xVals, yVals)
      xVals = xVals + x
      yVals = yVals + y
      call pgpoly(size(xVals), xVals, yVals)
    else
      call pgpt1(x, y, 2)
    end if

  end subroutine drawSrcEllipse

  !---------------------------------------------------------------------
  subroutine fixUpNulls(image, nonNullsArray)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout) :: image(:,:)
    logical, intent(out) :: nonNullsArray(:,:)

    real :: amin, amax

    !-------------------------------------------------------------------
    ! Start

    amin = minval(image)          ! seems to ignore null values
    amax = maxval(image) 

    ! It seems that no comparative returns T for NaNs, so I constructed
    ! this double-negative test that seems to work:

    nonNullsArray = .true. ! default
    where(.not.(image .ge. amin) .and. .not.(image .le. amax))
      nonNullsArray = .false.
      image = amin
    end where

  end subroutine fixUpNulls


!-----------------------------------------------------------------------
! set_palette: Sets a color palette for use by PGIMAG
!-----------------------------------------------------------------------
subroutine set_palette(itype)
integer, intent(in) :: itype    !colour palette: -1=inverse gray, 1=grey, 2=rainbow, 3=heat, 
                                ! 4=IRAF, 5=AIPS, 6=TJP
real, parameter :: al(20) = (/0.0, 0.1, 0.1, 0.2, 0.2, 0.3, 0.3, 0.4, 0.4, 0.5, &
                              0.5, 0.6, 0.6, 0.7, 0.7, 0.8, 0.8, 0.9, 0.9, 1.0/)
real, parameter :: ar(20) = (/0.0, 0.0, 0.3, 0.3, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0, &
                              0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0/)
real, parameter :: ag(20) = (/0.0, 0.0, 0.3, 0.3, 0.0, 0.0, 0.0, 0.0, 0.8, 0.8, &
                              0.6, 0.6, 1.0, 1.0, 1.0, 1.0, 0.8, 0.8, 0.0, 0.0/)
real, parameter :: ab(20) = (/0.0, 0.0, 0.3, 0.3, 0.7, 0.7, 0.7, 0.7, 0.9, 0.9, &
                              0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0/)
real, parameter :: grey(2) = (/0.0, 1.0/)
real, parameter :: invg(2) = (/1.0, 0.0/)
real, parameter :: hl(5) = (/0.0, 0.2, 0.4, 0.6, 1.0/)
real, parameter :: hr(5) = (/0.0, 0.5, 1.0, 1.0, 1.0/)
real, parameter :: hg(5) = (/0.0, 0.0, 0.5, 1.0, 1.0/)
real, parameter :: hb(5) = (/0.0, 0.0, 0.0, 0.3, 1.0/)
real, parameter :: rl(9) = (/-0.5, 0.0, 0.17, 0.33, 0.50, 0.67, 0.83, 1.0, 1.7/)
real, parameter :: rr(9) = (/0.0, 0.0, 0.0, 0.0, 0.6, 1.0, 1.0, 1.0, 1.0/)
real, parameter :: rg(9) = (/0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 0.6, 0.0, 1.0/)
real, parameter :: rb(9) = (/0.0, 0.3, 0.8, 1.0, 0.3, 0.0, 0.0, 0.0, 1.0/)
real, parameter :: tl(4) = (/0.0, 0.5, 0.5, 1.0/)
real, parameter :: tr(4) = (/0.2, 0.6, 0.6, 1.0/)
real, parameter :: tg(4) = (/0.0, 0.0, 0.5, 1.0/)
real, parameter :: tb(4) = (/1.0, 0.0, 0.0, 0.0/)
real, parameter :: wl(10) = (/0.0, 0.5, 0.5, 0.7, 0.7, 0.85, 0.85, 0.95, &
                              0.95, 1.0/)
real, parameter :: wr(10) = (/0.0, 1.0, 0.0, 0.0, 0.3,  0.8,  0.3,  1.0, &
                              1.0, 1.0/)
real, parameter :: wg(10) = (/0.0, 0.5, 0.4, 1.0, 0.0,  0.0,  0.2,  0.7,  &
                              1.0, 1.0/)
real, parameter :: wb(10) = (/0.0, 0.0, 0.0, 0.0, 0.4,  1.0,  0.0,  0.0, &
                              0.95, 1.0/)
! following designed by Clive Page: black + rainbow + white at top end
real, parameter :: cl(7) = (/0.1,  0.25, 0.43, 0.57, 0.71, 0.8,  1.0/)
real, parameter :: cr(7) = (/0.5,  0.0,  0.0,  0.0,  0.9,  1.0,  1.0/)
real, parameter :: cg(7) = (/0.0,  0.0,  0.8,  1.0,  0.9,  0.3,  1.0/)
real, parameter :: cb(7) = (/0.5,  1.0,  0.8,  0.0,  0.0,  0.3,  1.0/)
!                          black  mag   blue  cyan  grn   yell  red   white

select case(itype)
case(-2)                                             ! inverse rainbow
   call pgctab(rl, rr(9:1:-1), rg(9:1:-1), rb(9:1:-1), 9, 1.0, 0.5)
case(-1) 
   call pgctab(grey, invg, invg, invg, 2, 1.0, 0.5)  ! inverse greyscale
case(0)
call error('badColourMap', 'Sorry, --colourmap=0 is not allowed.')
case(1)
   call pgctab(grey, grey, grey, grey, 2, 1.0, 0.5)  ! greyscale
case(2)
   call pgctab(rl, rr, rg, rb, 9, 1.0, 0.5)          ! rainbow
case(3)
   call pgctab(hl, hr, hg, hb, 5, 1.0, 0.5)          ! heat
case(4)
   call pgctab(wl, wr, wg, wb, 10, 1.0, 0.5)         ! weird IRAF
case(5)
   call pgctab(al, ar, ag, ab, 20, 1.0, 0.5)         ! AIPS
case(6)
   call pgctab(tl, tr, tg, tb, 4, 1.0, 0.5)          ! Tim Pearson
case(7)
   call pgctab(cl, cr, cg, cb, 8, 1.0, 0.5)         ! Clive Page
   call pgscr(16, 0.0, 0.0, 0.0)                    ! bottom level is black
end select
return
end subroutine set_palette

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module implot_mod
