!+
! Module: wcs_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Returns semi-axes of an ellipse round each of a list of sources.
! 
! Comments: TODO:
! 
! $Id: wcs_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module wcs_aux

  use dal
  use errorHandling ! not needed now but should replace 'noisy' messages.

  use constants
  use coordinate	! for tanToRaDec.
  use fits_utils ! contains subs getRealKwd(), getIntKwd(), getStringKwd().

  implicit none 

  private 

  public getWcsInfo
  public easifyWcsType

  interface getWcsInfo
    module procedure getWcsInfo_image
    module procedure getWcsInfo_set
    module procedure getWcsInfo_table
  end interface

  interface easifyWcsType
    module procedure easifyWcsType_scalar
    module procedure easifyWcsType_vector
  end interface

  type, public :: wcsAxisType
    integer(int32) ::&
      refPixels
    character(len=8) ::&
      ctype
    real(double) ::&
      pixelSizeWorld,&
      refWorld
    integer(int8) ::&
      pixelSizeSign
  end type wcsAxisType

  type, public :: wcsDataType
    type(wcsAxisType) :: x, y
  end type wcsDataType

contains 

  !---------------------------------------------------------------------
  subroutine getWcsInfo_image(set, inArray, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set
    type(ArrayT),   intent(in) :: inArray
    type(wcsDataType), intent(out) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    call getIntKwd("CRPIX1",    set, inArray, wcs%X%refPixels)
    call getIntKwd("CRPIX2",    set, inArray, wcs%Y%refPixels)
    call getRealKwd("CDELT1",   set, inArray, wcs%X%pixelSizeWorld)
    call getRealKwd("CDELT2",   set, inArray, wcs%Y%pixelSizeWorld)
    call getRealKwd("CRVAL1",   set, inArray, wcs%X%refWorld)
    call getRealKwd("CRVAL2",   set, inArray, wcs%Y%refWorld)
    call getStringKwd("CTYPE1", set, inArray, wcs%X%ctype)
    call getStringKwd("CTYPE2", set, inArray, wcs%Y%ctype)

  end subroutine getWcsInfo_image

  !---------------------------------------------------------------------
  subroutine getWcsInfo_set(set, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set
    type(wcsDataType), intent(out) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    wcs%X%refPixels       = int32Attribute(set, 'CRPIX1')
    wcs%Y%refPixels       = int32Attribute(set, 'CRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(set, 'CDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(set, 'CDELT2')
    wcs%X%refWorld       = real32Attribute(set, 'CRVAL1')
    wcs%Y%refWorld       = real32Attribute(set, 'CRVAL2')
    wcs%X%ctype          = stringAttribute(set, 'CTYPE1')
    wcs%Y%ctype          = stringAttribute(set, 'CTYPE2')

  end subroutine getWcsInfo_set

  !---------------------------------------------------------------------
  subroutine getWcsInfo_table(tab, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    type(wcsDataType), intent(out) :: wcs

    !-------------------------------------------------------------------
    ! Start:

    wcs%X%refPixels       = int32Attribute(tab, 'CRPIX1')
    wcs%Y%refPixels       = int32Attribute(tab, 'CRPIX2')
    wcs%X%pixelSizeWorld = real32Attribute(tab, 'CDELT1')
    wcs%Y%pixelSizeWorld = real32Attribute(tab, 'CDELT2')
    wcs%X%refWorld       = real32Attribute(tab, 'CRVAL1')
    wcs%Y%refWorld       = real32Attribute(tab, 'CRVAL2')
    wcs%X%ctype          = stringAttribute(tab, 'CTYPE1')
    wcs%Y%ctype          = stringAttribute(tab, 'CTYPE2')

  end subroutine getWcsInfo_table

  !---------------------------------------------------------------------
  subroutine easifyWcsType_scalar(ctypeX, ctypeY, wcsType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: ctypeX, ctypeY
    character(8), intent(out) :: wcsType

    !-------------------------------------------------------------------
    ! Start:

    if (  trim(ctypeX) .eq. 'DETX' .and. trim(ctypeY) .eq. 'DETY') then
      wcsType = 'DETXY'

    else&
    if (trim(ctypeX).eq.'RA---TAN' .and. trim(ctypeY) .eq. 'DEC--TAN') then
      wcsType = 'XY'
    end if

  end subroutine easifyWcsType_scalar

  !---------------------------------------------------------------------
  subroutine easifyWcsType_vector(ctype, wcsType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: ctype(2)
    character(8), intent(out) :: wcsType

    !-------------------------------------------------------------------
    ! Start:

    if (  trim(ctype(1)) .eq. 'DETX' .and. trim(ctype(2)) .eq. 'DETY') then
      wcsType = 'DETXY'

    else&
    if (trim(ctype(1)).eq.'RA---TAN' .and. trim(ctype(2)) .eq. 'DEC--TAN') then
      wcsType = 'XY'
    end if

  end subroutine easifyWcsType_vector

  !---------------------------------------------------------------------
  function pixelsToTan(pixelValue, wcsAxis) result(tangentValue)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: pixelValue
    type(wcsAxisType), intent(in) :: wcsAxis
    real(single) :: tangentValue

    !-------------------------------------------------------------------
    ! Start

    tangentValue = (pixelValue - wcsAxis%refPixels)&
!    * wcsAxis%pixelSizeWorld * PiOn180
    * abs(wcsAxis%pixelSizeWorld) * PiOn180
!*** NOTE!! Use of the abs() here defines the tan system as always increasing with pixel index in both X and Y directions. This is NOT consistent with other uses of it in this library. At some point the rest should be brought into line.

! ...actually I would prefer also to define pixelSizeWorld always as positive.

  end function pixelsToTan

  !---------------------------------------------------------------------
  subroutine pixelXYtoRaDec(wcs, xi, yStart, yFinis&
  , raDegVector, decDegVector)
  !---------------------------------------------------------------------

    implicit none

    type(wcsDataType), intent(in) :: wcs
    integer(int16), intent(in) :: xi, yStart, yFinis
    real(double), intent(out) ::&
      raDegVector(yStart:yFinis), decDegVector(yStart:yFinis)

    integer(int16) :: yi
    real(double) ::&
      tanX(yStart:yFinis), tanY(yStart:yFinis)

    !-------------------------------------------------------------------
    ! Start

    tanX = (xi - 0.5 - wcs%x%refPixels) * abs(wcs%x%pixelSizeWorld) * PiOn180
    do yi = yStart, yFinis
      tanY(yi) = (yi - 0.5 - wcs%y%refPixels) * abs(wcs%y%pixelSizeWorld) * PiOn180
!print *, yi, tanX(yi), tanY(yi)
    end do
!print *, ''

    call tanToRaDec(wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180&
    , tanX, tanY, raDegVector, decDegVector)

    raDegVector  = raDegVector  / PiOn180
    decDegVector = decDegVector / PiOn180
    do yi = yStart, yFinis
!print *, yi, raDegVector(yi), decDegVector(yi)
    end do
!print *, ''

!print *, 'wcs%x%refPixels:',      wcs%x%refPixels
!print *, 'wcs%x%pixelSizeWorld:', wcs%x%pixelSizeWorld
!print *, 'wcs%x%refWorld:',       wcs%x%refWorld
!print *, 'wcs%y%refPixels:',      wcs%y%refPixels
!print *, 'wcs%y%pixelSizeWorld:', wcs%y%pixelSizeWorld
!print *, 'wcs%y%refWorld:',       wcs%y%refWorld

!call error('endPixelXYtoRaDec','')
  end subroutine pixelXYtoRaDec

!  !---------------------------------------------------------------------
!  subroutine worldToPixelCoords_scalar(wcs, worldvalue, pixelValue)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    type(wcsDataType),   intent(in)  :: wcs
!    type(srcRegionType), intent(in)  :: worldvalue
!    type(srcRegionType), intent(out) :: pixelValue
!
!    type(srcRegionType) :: worldvalueVector(1)
!    type(srcRegionType) :: pixelValueVector(1)
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    worldvalueVector(1) = worldvalue
!
!    call worldToPixelCoords_vector(wcs, worldvalueVector, pixelValueVector)
!
!    pixelValue = pixelValueVector(1)
!
!  end subroutine worldToPixelCoords_scalar
!
!  !---------------------------------------------------------------------
!  subroutine worldToPixelCoords_vector(wcs, srcDataWorld, srcDataPixels)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    type(wcsDataType),   intent(in)  :: wcs
!    type(srcRegionType), intent(in)  :: srcDataWorld(:)
!    type(srcRegionType), intent(out) :: srcDataPixels(:)
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    call message(AppMsg, Noisy, 'ssclib: convert world to pixel coords.')
!
!    srcDataPixels%x = (srcDataWorld%x - wcs%X%refWorld)&
!    / wcs%X%pixelSizeWorld + wcs%X%refPixels
!    srcDataPixels%y = (srcDataWorld%y - wcs%Y%refWorld)&
!    / wcs%Y%pixelSizeWorld + wcs%Y%refPixels
!!*** add 0.5 later for plotting purposes? (not for the region???)
!
!  end subroutine worldToPixelCoords_vector

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module wcs_aux
